'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _class, _temp;

const Promise = require('bluebird');
const Mservice = require('mservice');
const Mailer = require('ms-mailer-client');
const Errors = require('common-errors');
const merge = require('lodash/merge');
const fsort = require('redis-filtered-sort');
const TokenManager = require('ms-token');
const LockManager = require('dlock');
const defaultOpts = require('./defaults.js');
const RedisCluster = require('ioredis').Cluster;
const Flakeless = require('ms-flakeless');

const { NotImplementedError } = Errors;

/**
 * @namespace Users
 */
module.exports = (_temp = _class = class Users extends Mservice {

  /**
   * @namespace Users
   * @param  {Object} opts
   * @return {Users}
   */
  constructor(opts = {}) {
    super(merge({}, Users.defaultOpts, opts));
    this.initAdminAccounts = require('./accounts/init-admin.js');
    this.initFakeAccounts = require('./accounts/init-dev.js');
    const config = this.config;

    // id generator
    this.flake = new Flakeless(config.flake);

    this.on('plugin:connect:amqp', amqp => {
      this._mailer = new Mailer(amqp, config.mailer);
    });

    this.on('plugin:close:amqp', () => {
      this._mailer = null;
    });

    this.on('plugin:connect:redisCluster', redis => {
      fsort.attach(redis, 'fsort');

      // init token manager
      const tokenManagerOpts = { backend: { connection: redis } };
      this.tokenManager = new TokenManager(merge({}, config.tokenManager, tokenManagerOpts));

      // run migrations
      this.migrate('redis', `${__dirname}/migrations`);
    });

    // cleanup connections
    this.on('plugin:close:redisCluster', () => {
      this.dlock = null;
      this.tokenManager = null;
    });
  }

  /**
   * Getter for mailer client
   * @return {Object}
   */


  /**
   * Configuration options for the service
   * @type {Object}
   */
  get mailer() {
    const mailer = this._mailer;
    return mailer || this.emit('error', new NotImplementedError('amqp is not connected'));
  }

  /**
   * Getter for configuration
   * @return {Object}
   */
  get config() {
    return this._config;
  }

  /**
   * Gracefully connect to cluster
   * @return {Promise}
   */
  connect() {
    const config = this.config;
    this._pubsub = new RedisCluster(config.redis.hosts, _extends({}, config.redis.options, {
      lazyConnect: true
    }));

    return Promise.join(super.connect(), this._pubsub.connect()).tap(() => {
      // lock manager
      this.dlock = new LockManager(_extends({}, config.lockManager, {
        client: this._redis,
        pubsub: this._pubsub,
        log: this.log
      }));
    });
  }

  /**
   * Gracefully disconnect from the cluster
   * @return {Promise}
   */
  close() {
    return Promise.join(super.close(), this._pubsub.quit().reflect());
  }

  /**
   * Initializes Admin accounts
   * @return {Promise}
   */


  /**
   * Initializes fake account for dev purposes
   * @return {Promise}
   */
}, _class.defaultOpts = defaultOpts, _temp);